# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.


from setuptools import setup

setup(
    name="ragged_inference",
    author="Facebook AI Research",
    version="0.0.0",
    packages=["ragged_inference"],
    install_requires=[],
    scripts=[],
    python_requires=">=3.6",
)
